/*
  Phasetracker (C) 2008 Kasper Pedersen

  buildscript
  settings   dwarf2
  mcu        atmega32
  project    phasetracker
  output     phasetracker

  compile-s    main
  compile    main
  assemble   acquire

  link
  makeflash
  makeee
  makesym
  extlist
  end
*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

volatile unsigned char capbuf[432];
volatile unsigned char capbuf_full;
volatile unsigned char capbuf_divider;

signed rvt[2]; //retval for vector calculator


volatile unsigned char txbuf[32];
volatile unsigned char txhead,txtail;

extern void calcvector(unsigned char mask);
extern void calcvectorH1(unsigned char mask);
extern void calcvectorH2(unsigned char mask);


/*From KKP's collection of ugliness:
  returns the angle of a vector, +/- 1023 counts.
  peak error is 4 degrees.
*/
signed vectorangle2(signed x, signed y)
{
  unsigned char flags;
  unsigned a;

  flags=0;
  if (y<0) { y=-y; flags=1; } //map to positive Y half.
  if (x<0) { x=-x; flags|=2; } //map to positive X half.
  //map to 45 degree opening
  if (y>x) { x^=y; y^=x; x^=y; flags|=4; } //mirror around 45 degree ray
  //x is now guaranteed larger than y!
  while (y>0x7F) { x>>=1; y>>=1; }
  //convert to 8 bit fraction
  y<<=8;
  a=y/x; //max 0xFF result.
  if (flags&4) a=512-a;
  if (flags&2) a=1024-a;
  if (flags&1) a=-a;
  return a;
}











void ready_uart(void) //wait for uart to be ready for 16 characters again.
{
  unsigned used;
  do {
	used=(txhead-txtail)&31;
  } while (used>=15);
}

void putc(char ctx) //wait for uart to be ready for 32 characters again.
{
  unsigned char c;
  c=txhead;
  txbuf[c]=ctx;
  txhead=(c+1)&31;
  UCSRB|=(1<<UDRIE);
}

void ud5(unsigned v)
{
  putc(48+(v/10000)); v%=10000;
  putc(48+(v/1000)); v%=1000;
  putc(48+(v/100)); v%=100;
  putc(48+(v/10)); v%=10;
  putc(48+v);
}

void sd5(signed k)
{
  ready_uart();
  if (k>=0) {putc(' '); ud5( k);}
  else      {putc('-'); ud5(-k);}
}

void ud4(unsigned v)
{
  putc(48+(v/1000)); v%=1000;
  putc(48+(v/100)); v%=100;
  putc(48+(v/10)); v%=10;
  putc(48+v);
}

void sd4(signed k)
{
  ready_uart();
  if (k>=0) {putc(' '); ud4( k);}
  else      {putc('-'); ud4(-k);}
}

void hex4(unsigned v)
{
  char d;
  d=((v>>12)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=((v>>8)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=((v>>4)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=( v    &15)+48; if (d>57) d+=(65-48-10);
  putc(d);
}

static void hex3(unsigned v)
{
  char d;
  d=((v>>8)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=((v>>4)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=( v    &15)+48; if (d>57) d+=(65-48-10);
  putc(d);
}

static void hex2(unsigned char v)
{
  char d;
  d=((v>>4)&15)+48; if (d>57) d+=(65-48-10);
  putc(d);
  d=( v    &15)+48; if (d>57) d+=(65-48-10);
  putc(d);
}


void udpair_dec(unsigned char mask)
{
  calcvector(mask);
  sd5(rvt[0]);
  putc(' ');
  sd5(rvt[1]);
  putc(' ');
}

void udpair(unsigned char mask)
{
  calcvector(mask);
  ready_uart();
  hex4(rvt[0]);
  hex4(rvt[1]);
}

ISR(SIG_USART_DATA)
{
   unsigned char c=txtail;
   if (c==txhead) { //empty
	 UCSRB&=~(1<<UDRIE);
	 return;
   }
   UDR=txbuf[c];
   txtail=(c+1)&31;
}


int finefrq;
int coarsefrq(unsigned char mask)
{
  int p1,p2,dp,lastp;
  int sumx,sumy;
  long dps,finep;
  unsigned char ct;
  //approx 4000 clocks sample, 4000 clocks calculate, say, 10k clocks.
  //=about 1ms, 30ns max, 30ns/1ms=30 ppm measurement range.
  //
  //Inside a 3500 clock shot we have a distance of ~1600 clocks,
  //150 usec. 30ns/144us=207 ppm.
  //We can do coarse adjust on single shot to about 2ppm, then use high resolution.
  capbuf_full=0;
  ct=80;
  dps=0;
  finep=0;
  do {
	while (!capbuf_full) ;
	calcvectorH1(mask);
	sumx=rvt[0];
	sumy=rvt[1];
	p1=vectorangle2(sumx,sumy);
	calcvectorH2(mask);
	sumx+=rvt[0];
	sumy+=rvt[1];
	p2=vectorangle2(rvt[0],rvt[1]);
	dp=p2-p1;
	while (dp<-1024) dp+=2048;
	while (dp> 1024) dp-=2048;
	dps+=dp;

	p1=vectorangle2(sumx,sumy); //this time slot vector
	dp=p1-lastp; //compiler complains about noinit: dp gets nulled first run, OK.
	lastp=p1;
	if (ct==80) dp=0;
	while (dp<-1024) dp+=2048;
	while (dp> 1024) dp-=2048;
	finep+=dp;

	capbuf_full=0;
	capbuf_divider=1;
	//216*8 clocks center to center=1728ck=156us. 100ns=640 ppm.
	//we get 2048 counts for 640  ppm, or 0.3125 ppm/count
	//we have 3.2 count per ppm, to get 256 we add up 80 samples.
  } while (--ct);
  //finep is the phase skew over 79*5 ms=395 ms. 2048 ct/100ns, res=1.236E-10.
  //8.08 count per ppb. we accept that we show 1% too much error.
  finefrq=finep/8;
  dps>>=8;//div coarse.
  return dps;
}








void run_adjust(void)
{
  //output coarse and fine frequency. fine has +/-10ppm span and
  //will alias into this range if the actual frequency is off by more than that.
  //THUS: Adjust coarse to better than 5ppm, then look at fine.
  sd4(coarsefrq(1<<7)); //PPM error. calibration input on input 7.
  putc(' ');
  sd4(finefrq);         //PPB error
  putc(13);
  putc(10);
}

void run_phase(void)
{
	 while (!capbuf_full) ;
	 capbuf_full=0;
	 udpair(1);
	 udpair(2);
	 udpair(4);
	 udpair(8);
	 udpair(16);
	 udpair(32);
	 udpair(64);
	 udpair(128);
	 ready_uart();
	 putc(13);
	 putc(10);
}

void adconv(unsigned char n)
{
  ADMUX=(ADMUX&0xC0)|n;
  ADCSRA=0xC6;
  do {} while (ADCSRA&0x40);
  ready_uart();
  hex3(ADC&0x3FF);
}

void run_phase_ad(void)
{
  ADMUX=0xC0;  //internal reference, channel 0
  ADCSRA=0xC6; //ck/64=172kHz, T=864 clocks.

  while (!capbuf_full) ;
////  capbuf_divider=4;        //half rate = 50/second
  capbuf_full=0;
  udpair(1);
  udpair(2);
  udpair(4);
  udpair(8);
  udpair(16);
  udpair(32);
  udpair(64);
  udpair(128);  //8*8=64
  adconv(0);              //adconv takes 64*13.5=864 clocks=9 bit times.
  adconv(1);
  adconv(2);
  adconv(3);
  adconv(4);
  adconv(5);
  adconv(6);
  adconv(7);            //8*3=24 =>88 ch tot = 76% bandwidth.
  ready_uart();
  hex2(PINA);
  hex2(PINB);
  hex2(PINC);
  hex2(PIND);  //4*2=8 => 96 ch tot = 83.3%
  ready_uart();
  putc(13);
  putc(10);     //+2=>98 = 85.1%
}




void main(void)
{
   unsigned char n;

   PORTD|=(1<<2); //pullup
   PORTD|=(1<<3);

   OCR1A=55296-1;
   TCNT1=0;
   TCCR1A=0;
   TCCR1B=(1<<WGM12)|(1<<CS10); //start timer
   TIMSK=(1<<OCIE1A);
   capbuf_divider=1;

   UBRRL=(11059200/115200/16)-1;
   UCSRB=(1<<RXEN)|(1<<TXEN);
   UCSRC=(1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0); //N81 mode
   sei();

   //delay 5 ms
   while (!capbuf_full) ;
   capbuf_full=0;
   //the other end's uart should now be in sync.

   //calibration mode code.
   n=PIND&0x0C;
   if (n==0x0C) { //both are high.
	 PORTD&=~(1<<3); //drive D3 low.
	 DDRD|=(1<<3);
	 while (!capbuf_full) ;
	 capbuf_full=0;
	 while ((PIND&0x0C)==0x00) { //D2 is connected to D3 and NOT gnd.
	   run_adjust();
	 }
	 PORTD|=(1<<3);
	 DDRD&=~(1<<3);
   }

   for (;;) {
	 n=PIND&0x0C;
	 if (n==0x04) { //01
	   ADMUX=0x40;
	   run_phase_ad(); //AVCC-external-capacitor reference for ADC
	 } else
	 if (n==0x08) { //10
	   ADMUX=0xC0;
	   run_phase_ad(); //2.56V-internal reference for ADC
	 } else
	 if (n==0x0C) { //11
	   run_phase();
	 } else
	 { //00
	   ADMUX=0x00;
	   run_phase_ad(); //external reference for ADC
	 }
   }
}

